%%% -*- coding: utf-8 -*-
%%% -*- erlang-indent-level: 2 -*-
%%% -------------------------------------------------------------------
%%% Copyright 2010-2019 Manolis Papadakis <manopapad@gmail.com>,
%%%                     Eirini Arvaniti <eirinibob@gmail.com>
%%%                 and Kostis Sagonas <kostis@cs.ntua.gr>
%%%
%%% This file is part of PropEr.
%%%
%%% PropEr is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% PropEr is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with PropEr.  If not, see <http://www.gnu.org/licenses/>.

%%% @copyright 2010-2019 Manolis Papadakis, Eirini Arvaniti and Kostis Sagonas
%%% @version {@version}
%%% @author Manolis Papadakis
%%% @doc This module contains types for testing the typeserver.

-module(types_test2).
-export([function_that_uses_rec1/1, function_whose_spec_has_map_type/1]).

-export_type([exp1/1, exp2/0]).

-type exp1(T) :: {'a' | 'b', binary()} | {'c', T}.
-type exp2() :: atom() | [types_test1:exp1()].
-record(rec1, {f :: exp1(fun(() -> integer())), g :: fun((_,_) -> float())}).

function_that_uses_rec1(#rec1{}) ->
    ok.

%%
%% Added on the 18th of June 2019 to test the fix of issue #194
%%
-spec function_whose_spec_has_map_type(map()) -> map().
function_whose_spec_has_map_type(Map) ->
  Map#{key => 42}.
