%%% -*- coding: utf-8 -*-
%%% -*- erlang-indent-level: 2 -*-
%%% -------------------------------------------------------------------
%%% Copyright 2010-2019 Manolis Papadakis <manopapad@gmail.com>,
%%%                     Eirini Arvaniti <eirinibob@gmail.com>
%%%                 and Kostis Sagonas <kostis@cs.ntua.gr>
%%%
%%% This file is part of PropEr.
%%%
%%% PropEr is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% PropEr is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with PropEr.  If not, see <http://www.gnu.org/licenses/>.

%%% @copyright 2010-2019 Manolis Papadakis, Eirini Arvaniti and Kostis Sagonas
%%% @version {@version}
%%% @author Manolis Papadakis
%%% @doc This module tests whether native types are parsed outside of ?FORALLs.

-module(let_tests).
-export([]).

%% NOTE: Possibly here temporarily until the compiler's warnings are fixed.
-export_type([rng/0]).

-include_lib("proper/include/proper.hrl").

foo_t() ->
    ?LET(X, types_test1:exp1(), {42, X}).

prop_1() ->
    ?FORALL(_, foo_t(), true).

-type rng() :: 17..54.
bar_t() ->
    ?LET(N, rng(), 3*N).

prop_2() ->
    ?FORALL(N, bar_t(), N < 124).
